#!/usr/bin/env python3
"""
Quicksaver, published on the Mad Tea Party.

Will save designated save files, and keep set number of saves.
Retore the most recent on command.

Version: 2023.7.10
"""

import sys
sys.dont_write_bytecode = True

import json
import os
import shutil
import threading
import time
import unicodedata

###CONFIG STARTS###

SOURCE_LIST = [
	"/path/to/save/"
]

SAVE_PATH = os.path.join(os.getcwd(), "SAVES")

SAVE_DIR_NAME = "SAVE"
SAVE_META_NAME = "qsmeta.json"

LEGACY_SAVES_EXT = "_old"
LEGACY_SAVES_COUNT = 9

AUTOSAVE_PATH = os.path.join(os.getcwd(), "AUTOSAVES")

#interval is in seconds
AUTOSAVE_INTERVAL = 5
AUTOSAVE_DIR_NAME = "AUTOSAVE"
AUTOSAVE_META_NAME = SAVE_META_NAME

LEGACY_AUTOSAVES_EXT = LEGACY_SAVES_EXT
LEGACY_AUTOSAVES_COUNT = LEGACY_SAVES_COUNT

###CONFIG ENDS###

SOURCE_LIST = [pathlib.Path(i) for i in SOURCE_LIST]
if not isinstance(SAVE_PATH, pathlib.Path):
	SAVE_PATH = pathlib.Path(SAVE_PATH)
if not isinstance(AUTOSAVE_PATH, pathlib.Path):
	AUTOSAVE_PATH = pathlib.Path(AUTOSAVE_PATH)

def savedpl(p, e, n, d, j):
	# save dir to path while keeping legacy files
	# path, ext, legacy numbers, dir name, json name
	def savepl_inner(i=0):
		nonlocal p, e, n, d
		zp = len(str(n))
		if p.joinpath(d + e + str(i+1).zfill(zp)).exists():
			savepl_inner(i=(i+1))
		#leading zero padding
		if i == n:
			shutil.rmtree(str(p.joinpath(d + e + str(i).zfill(zp))))
		elif i == 0:
			p.joinpath(d).rename(p.joinpath(d + e + str(i+1).zfill(zp)))
		elif i < n:
			p.joinpath(d + e + str(i).zfill(zp)).rename(
				p.joinpath(d + e + str(i+1).zfill(zp)))
	if p.joinpath(d).exists():
		savepl_inner()
	qs = {}
	for i in SOURCE_LIST:
		dp = p.joinpath(d, i.name)
		shutil.copytree(i, str(dp))
		qs.update({str(dp): str(i)})
	with p.joinpath(d, j).open('w') as qsf:
		qsf.write(unicodedata.normalize('NFC',
			json.dumps(qs, 
				indent='\t', 
				sort_keys=True, 
				ensure_ascii=False)))

def save_auto_thread():
	global recentRestored
	while not es.is_set():
		if recentRestored:
			print("Recent restoration found. Skipping first autosave.")
			print("Will resume in " + str(AUTOSAVE_INTERVAL) + " seconds.\n")
			recentRestored = False
		else:
			savedpl(
				AUTOSAVE_PATH, 
				LEGACY_AUTOSAVES_EXT,
				LEGACY_AUTOSAVES_COUNT,
				AUTOSAVE_DIR_NAME, 
				AUTOSAVE_META_NAME)
			print("Autosave complete: " 
				+ time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()))
			print("Will resume in " + str(AUTOSAVE_INTERVAL) + " seconds.")
			print("Press [c] to interrupt autosave.\n")
		es.wait(timeout=AUTOSAVE_INTERVAL)

def restore_all_json(p, d, j, sp):
	# path, dir name, json name, save path
	with p.joinpath(d, j).open() as qsf:
		qs = json.load(qsf)
	for i in qs:
		shutil.copytree(
			str(p.joinpath(
				sp, 
				*(pathlib.Path(i).relative_to(p).parts[1:]))), 
			qs[i], 
			dirs_exist_ok=True)

if __name__=="__main__":
	global recentRestored
	recentRestored = False
	while True:
		match k := input("Press a key: [s]ave, [a]utosave, [r]estore or [q]uit.\n"):
			case 's':
				savedpl(
					SAVE_PATH,
					LEGACY_SAVES_EXT, 
					LEGACY_SAVES_COUNT, 
					SAVE_DIR_NAME, 
					SAVE_META_NAME)
				print("Save complete: " 
					+ time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()) + "\n")
			case 'a':
				t = threading.Thread(target=save_auto_thread)
				es = threading.Event()
				t.start()
				while True:
					ka_input = input()
					if ka_input == 'c':
						es.set()
						break
					else:
						print("Unidentified input: " + ka_input)
				t.join()
			case 'r':
				while True:
					match kf := input("Restore from: [s]ave, [a]utosave, or [c] to interrupt.\n"):
						case 's':
							p = SAVE_PATH
							d = SAVE_DIR_NAME
							e = LEGACY_SAVES_EXT
							j = SAVE_META_NAME
							n = LEGACY_SAVES_COUNT
							break
						case 'a':
							p = AUTOSAVE_PATH
							d = AUTOSAVE_DIR_NAME
							e = LEGACY_AUTOSAVES_EXT
							j = AUTOSAVE_META_NAME
							n = LEGACY_AUTOSAVES_COUNT
							break
						case 'c':
							break
						case _:
							print("Unidentified input: " + kf)
				while True and (kf != 'c'):
					match ke := input("Enter the extension number, 0 for most recent, or [c] to interrupt.\n"):
						case 'c':
							break
						case '0':
							sp = d
						case _:
							sp = d + e + str(ke).zfill(len(str(n)))
					if p.joinpath(sp).exists():
						break
					else:
						print("Save not found: " + sp)
				if (kf != 'c') and (ke != 'c'):
					restore_all_json(p, d, j, sp)
					recentRestored = True
					print("Restored: " + sp)
			case 'q':
				exit()
			case _: 
				print("Unidentified input: " + k)